@@ -5,7 +5,7 @@ use warnings;
 use DBI;
 use Carp qw(croak);
 
-our $VERSION=0.02;
+our $VERSION=0.03;
 
 =head1 NAME
 
@@ -74,7 +74,8 @@ be referred to as a node.
 
 =head3 new
 
-    $root_node = DBIx::Path->new(%config);
+    $root_node = DBIx::Path->new( %config );
+    $node = $root_node->new( id => $id );    # If the ID is known
 
 Manually creates an object to represent the tree's root node.  Note that there doesn't 
 actually need to be an ID in the table for this node, although it doesn't hurt!  The 
@@ -82,6 +83,9 @@ C<pid> and C<name> methods will always return C<undef> for a node created throug
 but nodes created through C<new> otherwise behave identically to those created with 
 C<get>, C<set>, C<add>, or C<resolve>.
 
+C<new> can also be used as a copy constructor.  This can be used to create a DBIx::Path
+object from a known ID.
+
 The arguments comprise a hash (not a hash reference) with the following keys:
 
 =over 4
@@ -112,6 +116,27 @@ The name of the name field.  Optional; defaults to "name".
 The ID of the root.  Optional; defaults to 0.  Note that C<id> 
 cannot be C<undef> (C<NULL>), due to the SQL used to retrieve nodes.
 
+=item * C<hooks>
+
+A hashref of sub references which C<DBIx::Path> will call at appropriate
+time.  Optional; current hooks include:
+
+=over 4
+
+=item C<lock>
+
+Called at the beginning of any operation involving a database query.  The first
+parameter is the database handle; the second is 'r' for a read operation (which
+recommends a shared lock) or 'w' for a write operation (which recommends an
+exclusive lock).
+
+=item C<unlock>
+
+Called after the operation which needed a lock is complete.  Parameters are the
+same as C<lock>.
+
+=back 4
+
 =back 4
 
 The return value is a DBIx::Path object, which can then be used normally.
@@ -155,6 +180,11 @@ END
 SELECT		$config{id_column}, $config{pid_column}, $config{name_column}
 	FROM	$config{table} 
 	WHERE	$config{pid_column} = ?
+END
+        $config{sth}{parents}=$dbh->prepare(<<"END") or die "DBIx::Path->new: couldn't prepare 'parents' statement: $DBI::errstr";
+SELECT      $config{id_column}, $config{pid_column}, $config{name_column}
+    FROM    $config{table}
+    WHERE   $config{id_column} = ?
 END
     }
 
@@ -176,7 +206,9 @@ exists.
 
 sub get {
 	my($me, $name)=@_;
+    $me->_lock('r');
 	$me->{sth}{get}->execute($me->{id}, $name);
+    $me->_unlock('r');
 	my $r=$me->{sth}{get}->fetchrow_arrayref() or do {
 		require POSIX;
 		$!=POSIX::ENOENT();
@@ -201,6 +233,7 @@ sub add {
 	my($me, $name, $id)=@_;
 	$id = $id->id if ref $id;    #Be nice to the poor, dumb hacker.
 
+    $me->_lock('w');
     #Make sure such a row doesn't exist first.
 	my $affected=(
         ! $me->get($name) &&
@@ -208,12 +241,15 @@ sub add {
     );
     
     if(!defined $affected or $affected == 0) {
+        $me->_unlock('w');
 		require POSIX;
 		$!=POSIX::EEXIST();
 		return;
 	}
     else {
-	    return $me->get($name);
+	    my $ret=$me->get($name);
+        $me->_unlock('w');
+        return $ret;
 	}
 }
 
@@ -229,7 +265,9 @@ If the SQL statement failed, returns C<undef> and sets $! to ENOENT.
 sub del {
 	my($me, $name)=@_;
     
+    $me->_lock('w');
     my $affected=$me->{sth}{del}->execute($me->{id}, $name);
+    $me->_unlock('w');
 
     if(!defined $affected or $affected == 0) {
 		require POSIX;
@@ -250,39 +288,68 @@ making $id the child of $node named $name.  A simple wrapper around
 C<del> and C<add>.  Return values are the same as C<add>, but note that
 $! may still be set due to the result of C<del>.
 
-This method may be subject to race conditions.  Patches to fix this problem
-in a cross-database manner are welcome.
+This method may be subject to race conditions unless the C<lock> and
+C<unlock> hooks are specified.
 
 =cut
 
 sub set {
 	my($me, $name, $id)=@_;
+    $me->_lock('w');
 	$me->del($name);
-	goto &add;
+	my $ret=$me->add($name, $id);
+    $me->_unlock('w');
+    return $ret;
 }
 
+=head3 parents
+
+    @list = $node->parents()
+
+Finds all references to the current node--in other words, all parents which count this
+node as their child.  In scalar context, returns the number of parents; in list context,
+returns a hash whose keys are the IDs of the parents and whose values are an arrayref of
+names the current node is found under.
+
+=cut
+
+sub parents {
+    my($me)=@_;
+    my %ret;
+
+    $me->_lock('r');
+    $me->{sth}{parents}->execute($me->id);
+    $me->_unlock('r');
+    while(my $r=$me->{sth}{parents}->fetchrow_arrayref) {
+        push @{$ret{$r->[1]}}, $r->[2];
+    }
+
+    return wantarray ? %ret : scalar keys %ret;
+}
+
+
 =head3 resolve
 
     $node = $node->resolve( @components )
 
-The C<resolve> method traverses the provided path; that is, it looks 
-up the child of $node named $components[0], then looks up the child of 
+The C<resolve> method traverses the provided path; that is, it looks
+up the child of $node named $components[0], then looks up the child of
 the just-retrieved node named $components[1], and so on.  Components
 which are undefined or consist of the empty string will be ignored.
 
-Return value is the same as C<get> when a name anywhere in @components 
+Return value is the same as C<get> when a name anywhere in @components
 does not resolve.
 
-After it is run, @DBIx::Path::RESOLVED will contain all components it 
-resolved, and $DBIx::Path::PARENT will contain the node resolved from 
-$RESOLVED[-1].  That is, after a successful run, @RESOLVED will be a 
-copy of the @components list, and $PARENT will contain the parent of 
+After it is run, @DBIx::Path::RESOLVED will contain all components it
+resolved, and $DBIx::Path::PARENT will contain the node resolved from
+$RESOLVED[-1].  That is, after a successful run, @RESOLVED will be a
+copy of the @components list, and $PARENT will contain the parent of
 the returned node.
 
-After a failed attempt to resolve, @RESOLVED will contain all components 
+After a failed attempt to resolve, @RESOLVED will contain all components
 that resolved successfully.  An additional variable, @DBIx::Path::FAILED,
-will contain the remaining components.  $PARENT will contain the node which 
-didn't have a child named $FAILED[0].  These variables are intended to augment 
+will contain the remaining components.  $PARENT will contain the node which
+didn't have a child named $FAILED[0].  These variables are intended to augment
 the simple ENOENT placed in $! by C<get>.
 
 =cut
@@ -296,18 +363,96 @@ sub resolve {
 	our @FAILED=@components;
 	our $PARENT=$me;
 	
+    $me->_lock('r');
 	for(@components) {
         next unless defined $_ and $_ ne '';
 		$PARENT=$cursor;
 		$cursor=$cursor->get($_);
-		return undef unless defined $cursor;
+		last unless defined $cursor;
 		push @RESOLVED, $_;
 		shift @FAILED;
 	}
+    $me->_unlock('r');
 	
 	return $cursor;
 }
 
+=head3 reverse
+
+    @list=$node->reverse()
+
+Returns all paths from the root to the current node.  The return value is an
+arrayref of names.  (If called on the root node, it returns a single empty 
+arrayref, which is probably the most accurate depiction of the situation.)
+If called in scalar context, it will return the number of paths from the root to
+the current node.
+
+If reverse() encounters a circular reference while attempting to find the parents,
+it will return an empty list and set $! to ELOOP.
+
+The results of this function are undefined if the relevant parts of the tree are
+modified while it's being executed.  It is an exceedingly good idea to specify the
+C<lock> and C<unlock> hooks if you intend to use this method.
+
+Note: This function disregards the contents of the C<id> parameter given to C<new>; 
+it considers the root to simply be the parentless node at the top of the tree.  
+Thus, it may "break out" of a particular subtree.
+
+=cut
+
+{
+our %seen;
+my %memo;
+sub reverse {
+    %seen=();
+    %memo=();
+    $_[0]->_lock('r');
+    my @ret;
+    eval {
+        @ret=&_reverse;
+    };
+    $_[0]->_unlock('r');
+    if($@) {
+       if($@ =~ /Circular reference encountered during DBIx::Path->reverse()/) {
+            require POSIX;
+            $!=POSIX::ELOOP();
+            return;
+	    }
+        else {
+            die;
+        }
+    }
+    return @ret;
+}
+
+sub _reverse {
+    my($me)=@_;
+    my @ret;
+
+    local %seen=%seen;
+    die "Circular reference encountered during DBIx::Path->reverse()" if $seen{$me->id};
+    $seen{$me->id}++;
+
+    return @{$memo{$me->id}} if $memo{$me->id};
+
+    my %parents=$me->parents;
+    return [] unless %parents;
+
+    for my $pid(keys %parents) {
+        my $parent=$me->new(%$me, id => $pid);
+        my @parent_ret=$parent->_reverse();
+        for my $name(@{ $parents{$pid} }) {
+			 for my $ancestors(@parent_ret) {
+                 push @ret, [@$ancestors, $name];
+             }
+        }
+    }
+    $memo{$me->id}=\@ret;
+    return @ret;
+}
+}
+
+
 =head3 list
 
     @nodes=$node->list()
@@ -324,7 +469,9 @@ sub list {
 	my @ret;
 	local $_;
 	
+    $me->_lock('r');
 	$me->{sth}{list}->execute($me->{id});
+    $me->_unlock('r');
 	push @ret, $me->_row_to_obj($_) while $_ = $me->{sth}{list}->fetchrow_arrayref;
 	return @ret;
 }
@@ -358,6 +505,32 @@ for my $field qw(id pid name) {
 	*{$field}=sub { $_[0]->{$field} }
 }
 
+{
+    my $locked=0;
+    sub _lock {
+        my($me, $type)=@_;
+        $locked++;
+        if($locked == 1 and $me->{hooks}{lock}) {
+            $me->{hooks}{lock}->($me->{dbh}, $type);
+ 	    }
+    }
+    sub _unlock {
+        my($me, $type)=@_;
+        if($locked == 1) {
+            $me->{hooks}{unlock}->($me->{dbh}, $type) if $me->{hooks}{unlock};
+		}
+        elsif($locked == 0) {
+            croak "DBIx::Path: PANIC: Key won't fit in lock";
+		}
+        $locked--;
+    }
+    END {
+	    if($locked) {
+            warn "DBIx::Path: WARNING: Program may have exited with lock(s) still held";
+		}
+    }
+}
+
 sub _row_to_obj {
 	my($id, $pid, $name)=@{$_[1]};
 	(ref $_[0])->new(%{$_[0]}, id => $id, pid => $pid, name => $name);
@@ -381,7 +554,7 @@ table name (such as an C<undef> value) was passed.  Please check your code.
 
 C<new> prepares several SQL statements which are used by the other
 methods.  This message indicates that the indicated statement was invalid.
-	This could indicate a bad table name or invalid I<whatever>_column settings;
+This could indicate a bad table name or invalid I<whatever>_column settings;
 it could also mean that the SQL used by DBIx::Path isn't recognized by your
 DBD.
 
@@ -389,32 +562,60 @@ Check the parameters you're passing to DBIx::Path->new, then make sure the
 SQL at the indicated line number is valid for your server.  The text after
 the second colon is the DBI error message.
 
+=item C<DBIx::Path: PANIC: Key won't fit in lock>
+
+Internal error indicating that DBIx::Path wanted to call your unlock hook,
+but somehow forgot to call your lock hook in the first place.  This shouldn't
+happen.
+
+=item C<DBIx::Path: WARNING: Program may have exited with lock(s) still held>
+
+DBIx::Path has an internal variable tracking how deeply nested locks currently
+are; it uses this to ensure that your lock/unlock hooks are only called once
+per method.  This message will be printed to STDERR if that variable has a 
+non-zero value when Perl exits; it indicates that you should check to make sure
+your database hasn't become wedged.
+
+This message is most often seen when an error occurs within DBIx::Path itself;
+it may also be displayed for particularly nasty sorts of bugs within DBIx::Path
+or if your process is hit by a signal while DBIx::Path is querying your 
+database.  It's never good news.
+
 =back 4
 
 =head1 BUGS AND ENHANCEMENTS
 
-The implementation of C<set> may be vulneurable to race conditions.  Other than
-that, there are no known bugs at this time; however, I'm not that experienced 
+There are no known bugs at this time; however, I'm not that experienced
 with the DBI, so God only knows if I've missed something important.
 
+The following methods may be vulneurable to race conditions or other time-sensitive
+bugs unless locking hooks are provided:
+
+=over 4
+
+=item * C<set>
+
+=item * C<resolve>
+
+=item * C<reverse>
+
+=back 4
+
 Some enhancements I'm considering are:
 
 =over 4
 
-=item * Hooks on the basic operations (C<get>, C<add>, C<del>, possibly C<set> 
+=item * Hooks on the basic operations (C<get>, C<add>, C<del>, possibly C<set>
 and C<list>).  Subclassing may make this unnecessary, however.
 
 =item * Methods that select all descendents of the current node and return them
 in various useful forms.  (These would have to not curl up into a ball and cry
 in the face of circular references.)
 
-=item * Reverse lookups--given two nodes, figure out how to get from one to the
-other.
-
 =back 4
 
-Patches to implement these, or to fix bugs, are much appreciated; send them to 
-<brentdax@cpan.org> and start the subject line with something like 
+Patches to implement these, or to fix bugs, are much appreciated; send them to
+<brentdax@cpan.org> and start the subject line with something like
 "[PATCH DBIx::Path]".
 
 =head1 SEE ALSO
@@ -429,7 +630,7 @@ Brent 'Dax' Royal-Gordon, <brentdax@cpan.org>
 
 Copyright 2005 Brent 'Dax' Royal-Gordon.  All rights reserved.
 
-This program is free software; it may be used, redistributed, and/or modified 
+This library is free software; it may be used, redistributed, and/or modified 
 under the same terms as Perl itself.
 
 =cut
